# -*- coding: utf-8 -*-
"""
Helper functions for handling JSON data from Ostriz.
"""

from __future__ import unicode_literals, absolute_import

import os
import io
import json
import datetime
import requests

from dump2polarion.exporter import ImportedData
from dump2polarion.exceptions import Dump2PolarionException


def _get_json(location):
    """Reads JSON data from file or URL."""
    try:
        json_data = requests.get(location)
    except ValueError:
        pass
    else:
        return json.loads(json_data.text).get('tests')

    location = os.path.expanduser(location)
    if os.path.isfile(location):
        with io.open(location, encoding='utf-8') as json_data:
            try:
                return json.load(json_data).get('tests')
            except Exception as err:
                raise Dump2PolarionException(
                    "Failed to parse JSON from {}: {}".format(location, err))


def _get_testrun_id(version):
    """Gets testrun id out of the appliance_version file."""
    try:
        build_base = version.strip().split('-')[0].split('_')[0].replace('.', '_')
        zval = int(build_base.split('_')[3])
    except Exception:
        # not in expected format
        raise Dump2PolarionException("Cannot find testrun id")
    if zval < 10:
        pad_build = build_base[-1].zfill(2)
        return build_base[:-1] + pad_build
    return build_base


def _calculate_duration(start_time, finish_time):
    """Calculates how long it took to execute the testcase."""
    if not(start_time and finish_time):
        return 0
    start = datetime.datetime.fromtimestamp(start_time)
    finish = datetime.datetime.fromtimestamp(finish_time)
    duration = finish - start
    return duration.seconds


def _parse_ostriz(ostriz_data):
    """Reads the content of the input JSON and returns testcases results."""
    if not ostriz_data:
        raise Dump2PolarionException("No data to import")
    test_data = []
    data = found_version = None
    for data in ostriz_data.values():
        # make sure we are collecting data for the same appliance version
        if found_version:
            if found_version != data.get('version'):
                continue
        else:
            found_version = data.get('version')
        statuses = data.get('statuses')
        if not statuses:
            continue
        test_data.append(dict(
            verdict=statuses.get('overall'),
            title=data.get('test_name'),
            time=_calculate_duration(data.get('start_time'), data.get('finish_time')) or 0))
    testrun_id = _get_testrun_id(found_version)
    return ImportedData(results=test_data, testrun=testrun_id)


# pylint: disable=unused-argument
def import_ostriz(location, **kwargs):
    """Reads Ostriz's data and returns imported data."""
    ostriz_data = _get_json(location)
    return _parse_ostriz(ostriz_data)



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Connects to the Polarion Importer message bus and verifies that results were submitted.
"""

from __future__ import unicode_literals, absolute_import

import os
import logging
import threading
import time
import json
import pprint

from xml.etree import ElementTree

import requests


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class _XunitListener(object):
    """Listener for xunit importer message bus."""
    def __init__(self):
        self.message_list = []
        self.message_condition = threading.Condition()
        self.message_received = False

    def on_message(self, headers, message, is_error=False):
        """Actions when message is received."""
        self.message_list.append((headers, message, is_error))
        with self.message_condition:
            self.message_received = True
            self.message_condition.notify()

    def on_error(self, headers, message):
        """Actions when error is received."""
        return self.on_message(headers, message, True)

    def wait_for_message(self, timeout=None):
        """Waits for message on xunit importer message bus."""
        timeout = timeout or 300
        with self.message_condition:
            if not self.message_received:
                self.message_condition.wait(timeout=timeout)
        retval = self.message_received
        self.message_received = False
        return retval

    def get_latest_message(self):
        """Returns last received message."""
        return self.message_list[-1]


def _get_response_property(xunit):
    """Parse xunit xml and finds the "polarion-response-" name and value."""
    try:
        root = ElementTree.fromstring(xunit.encode('utf-8'))
    except ElementTree.ParseError as err:
        logger.error(err)
        return

    if root.tag == 'testsuites':
        properties = root.find('properties')
        for prop in properties:
            prop_name = prop.get('name', '')
            if 'polarion-response-' in prop_name:
                return (prop_name[len('polarion-response-'):], str(prop.get('value')))
    elif root.tag == 'testcases':
        properties = root.find('response-properties')
        for prop in properties:
            if prop.tag != 'response-property':
                continue
            prop_name = prop.get('name')
            prop_value = prop.get('value')
            if prop_name and prop_value:
                return (prop_name, str(prop_value))


def _log_received_data(headers, message):
    """Logs received message and headers."""
    if not logger.isEnabledFor(logging.DEBUG) or headers is None or message is None:
        return
    pp = pprint.PrettyPrinter(indent=4)
    logging.debug("Message headers: \n{}".format(pp.pformat(headers)))
    logging.debug("Message body: \n{}".format(message))


def _download_log(url, output_file):
    """Saves log returned by the message bus."""
    logger.info("Saving log {} to {}".format(url, output_file))

    def _do_log_download():
        try:
            return requests.get(url)
        # pylint: disable=broad-except
        except Exception as err:
            logger.error(err)

    # log file may not be ready yet, wait a bit
    for _ in range(5):
        log_data = _do_log_download()
        if log_data or log_data is None:
            break
        time.sleep(2)

    if not log_data:
        logger.error("Failed to download log file '{}'.".format(url))
    with open(os.path.expanduser(output_file), 'wb') as out:
        out.write(log_data.content)


def _check_outcome(message, is_error, log_file=None):
    """Parses returned message and checks submit outcome."""
    if is_error is None:
        logger.error("Submit verification timed out, check in the web UI if results were updated "
                     "(might take a while, please don't resubmit)")
        return False
    elif is_error:
        logger.error("Received an error, results not updated")
        return False

    try:
        data = json.loads(message)
    # pylint: disable=broad-except
    except Exception:
        logger.error("Cannot parse message, results probably not updated")
        return False

    url = data.get('log-url')
    if url:
        if not log_file:
            logger.info("Submit log: {}".format(url))
        else:
            _download_log(url, log_file)

    if data.get('status') == 'passed':
        logger.info("Results successfully submitted!")
        return True

    logger.error("Status = {}, results not updated".format(data.get('status')))
    return False


def get_verification_func(bus_url, xunit, user, password, **kwargs):
    """Subscribes to the message bus and returns verification function."""
    if not bus_url:
        logger.error(
            "Message bus url ('message_bus') not configured, skipping submit verification")
        return

    selector = _get_response_property(xunit)
    if not selector:
        logger.error(
            "The response property is not set, skipping submit verification")
        return

    if not all([user, password]):
        logger.error("Missing credentials, skipping submit verification")
        return

    host, port = bus_url.split(':')

    # avoid slow initialization of stomp when it's not needed
    import stomp
    logging.getLogger('stomp.py').setLevel(logging.WARNING)
    conn = stomp.Connection([(host.encode('ascii', 'ignore'), int(port))])

    listener = _XunitListener()
    conn.set_listener('Importer Listener', listener)
    logger.debug("Subscribing to the Importer message bus")
    conn.start()
    conn.connect(login=user, passcode=password)

    try:
        conn.subscribe(
            destination='/topic/CI',
            id=1,
            ack='auto',
            headers={'selector': "{}='{}'".format(selector[0], selector[1])}
        )
    # pylint: disable=broad-except
    except Exception as err:
        logger.error("Skipping submit verification: {}".format(err))
        logger.debug("Terminating subscription")
        conn.disconnect()

    def verify_submit(skip=False, timeout=None):
        """Verifies that the results were successfully submitted."""
        headers = message = is_error = None
        try:
            if skip:
                # just do cleanup in finally
                return
            logger.info("Waiting for response on the Importer message bus...")
            logger.debug("Response selector: {}={}".format(selector[0], selector[1]))
            if listener.wait_for_message(timeout=timeout):
                headers, message, is_error = listener.get_latest_message()
        # pylint: disable=broad-except
        except Exception as err:
            logger.error("Skipping submit verification: {}".format(err))
        finally:
            logger.debug("Terminating subscription")
            conn.disconnect()

        _log_received_data(headers, message)

        return _check_outcome(message, is_error, log_file=kwargs.get('log_file'))

    return verify_submit



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Configuration loading.
"""

from __future__ import unicode_literals, absolute_import

import os
import io
import logging
import yaml

from dump2polarion.exceptions import Dump2PolarionException

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def get_config(config_file=None):
    """Loads config file and returns its content."""
    default_conf = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dump2polarion.yaml')
    user_conf = config_file or '~/.config/dump2polarion.yaml'

    try:
        with open(os.path.expanduser(user_conf)):
            pass
    except EnvironmentError:
        user_conf = None
        if config_file:
            raise Dump2PolarionException("Cannot open config file '{}'".format(config_file))

    with io.open(default_conf, encoding='utf-8') as input_file:
        config_settings = yaml.load(input_file)
    logger.debug("Default config loaded from '{}'".format(default_conf))

    if user_conf:
        with io.open(user_conf, encoding='utf-8') as input_file:
            config_settings_user = yaml.load(input_file)
        logger.info("Config loaded from '{}'".format(user_conf))

        # merge default and user configuration
        try:
            config_settings.update(config_settings_user)
        except ValueError as err:
            raise Dump2PolarionException(
                "Failed to load the '{}' config file: {}".format(user_conf, err))

    return config_settings



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Utils for dump2polarion.
"""

from __future__ import unicode_literals, absolute_import

import os
import io
import datetime
import string
import random
import logging

from xml.etree import ElementTree

from dump2polarion.exceptions import Dump2PolarionException


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def write_xml(xml, output_loc=None, filename=None):
    """Outputs the XML content into a file.

    Args:
        xml: string with XML document
        output_loc: file or directory for saving the file
        filename: file name that will be used if output_loc is directory
            If it is needed and is not supplied, it will be generated
    """
    if not xml:
        raise Dump2PolarionException("No data to write.")
    filename = filename or 'output-{}-{:%Y%m%d%H%M%S}.xml'.format(
        ''.join(random.sample(string.lowercase, 5)), datetime.datetime.now())
    if output_loc:
        filename_fin = os.path.expanduser(output_loc)
        if os.path.isdir(filename_fin):
            filename_fin = os.path.join(filename_fin, filename)
    else:
        filename_fin = filename

    with io.open(filename_fin, 'w', encoding='utf-8') as xml_file:
        xml_file.write(get_unicode_str(xml))
    logger.info("Data written to '{}'".format(filename_fin))


def xunit_fill_testrun_id(xml, testrun_id):
    """Adds the polarion-testrun-id property when it's missing."""
    try:
        xml_root = ElementTree.fromstring(xml.encode('utf-8'))
    # pylint: disable=broad-except
    except Exception as err:
        raise Dump2PolarionException("Failed to parse XML file: {}".format(err))

    properties = xml_root.find('properties')
    if properties is None:
        raise Dump2PolarionException("Failed to find <properties> in the XML file")
    for prop in properties:
        if prop.get('name') == 'polarion-testrun-id':
            return xml

    ElementTree.SubElement(properties, 'property',
                           {'name': 'polarion-testrun-id', 'value': str(testrun_id)})
    return get_unicode_str(ElementTree.tostring(xml_root, encoding='utf-8'))


def get_unicode_str(obj):
    """Makes sure obj is a unicode string."""
    if isinstance(obj, unicode):
        return obj
    if isinstance(obj, str):
        return obj.decode('utf-8', errors='ignore')
    return unicode(obj)


def init_log(log_level):
    """Initializes logging."""
    log_level = log_level or 'INFO'
    logging.basicConfig(
        format='%(name)s:%(levelname)s:%(message)s',
        level=getattr(logging, log_level.upper(), logging.INFO))



# -*- coding: utf-8 -*-
"""
Helper functions for handling data in CSV format.
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import csv

from collections import OrderedDict

from dump2polarion.exporter import ImportedData
from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.csv_unicode import UnicodeReader


def _get_csv_fieldnames(csv_reader):
    """Finds fieldnames in Polarion exported csv file."""
    fieldnames = []
    for row in csv_reader:
        for col in row:
            field = (col.
                     strip().
                     replace('"', '').
                     replace(' ', '').
                     replace('(', '').
                     replace(')', '').
                     lower())
            fieldnames.append(field)
        if 'id' in fieldnames:
            break
        else:
            # this is not a row with fieldnames
            del fieldnames[:]
    if not fieldnames:
        return
    # remove trailing unannotated fields
    while True:
        field = fieldnames.pop()
        if field:
            fieldnames.append(field)
            break
    # name unannotated fields
    suffix = 1
    for index, field in enumerate(fieldnames):
        if not field:
            fieldnames[index] = 'field{}'.format(suffix)
            suffix += 1

    return fieldnames


def _get_testrun_from_csv(file_obj, csv_reader):
    """Tries to find the testrun id in  Polarion exported csv file."""
    file_obj.seek(0)
    search_str = r'TEST_RECORDS:\("[^/]+/([^"]+)"'
    testrun_id = None
    too_far = False
    for row in csv_reader:
        for col in row:
            if not col:
                continue
            field = (col.
                     strip().
                     replace('"', '').
                     replace(' ', '').
                     lower())
            if field == 'id':
                # we are too far, tests results start here
                too_far = True
                break
            search = re.search(search_str, col)
            try:
                testrun_id = search.group(1)
            except AttributeError:
                continue
            else:
                break
        if testrun_id or too_far:
            break

    return testrun_id


def _get_results(csv_reader, fieldnames):
    """Maps data to fieldnames.

    The reader needs to be at position after fieldnames, before the results data.
    """
    fieldnames_count = len(fieldnames)
    results = []
    for row in csv_reader:
        for col in row:
            if col:
                break
        else:
            # empty row, skip it
            continue
        record = OrderedDict(zip(fieldnames, row))
        # skip rows that were already exported
        if record.get('exported') == 'yes':
            continue
        row_len = len(row)
        if fieldnames_count > row_len:
            for key in fieldnames[row_len:]:
                record[key] = None
        results.append(record)

    return results


def _get_csv_reader(input_file):
    """Returns csv reader."""
    dialect = csv.Sniffer().sniff(input_file.read(2048))
    input_file.seek(0)
    return UnicodeReader(input_file, dialect)


# pylint: disable=unused-argument
def get_imported_data(csv_file, **kwargs):
    """Reads the content of the Polarion exported csv file and returns imported data."""
    with open(os.path.expanduser(csv_file), 'rb') as input_file:
        reader = _get_csv_reader(input_file)

        fieldnames = _get_csv_fieldnames(reader)
        if not fieldnames:
            raise Dump2PolarionException(
                "Cannot find field names in CSV file '{}'".format(csv_file))

        results = _get_results(reader, fieldnames)
        if not results:
            raise Dump2PolarionException(
                "No results read from CSV file '{}'".format(csv_file))

        testrun = _get_testrun_from_csv(input_file, reader)

    return ImportedData(results=results, testrun=testrun)


def _check_required_columns(csv_file, results):
    required_columns = {'verdict': 'Verdict'}
    missing_columns = [required_columns[k] for k in required_columns if k not in results[0]]
    if missing_columns:
        raise Dump2PolarionException(
            "The input file '{}' is missing following columns: {}".format(
                csv_file, ', '.join(missing_columns)))


def import_csv(csv_file, **kwargs):
    """Imports data and checks that all required columns are there."""
    records = get_imported_data(csv_file, **kwargs)
    _check_required_columns(csv_file, records.results)
    return records



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Dump testcases results from a CSV input file to SQLite.
"""

from __future__ import unicode_literals, absolute_import

import argparse
import datetime
import logging
import os
import sqlite3

from dump2polarion import csvtools
from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.utils import init_log


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def get_args(args=None):
    """Get command line arguments."""
    parser = argparse.ArgumentParser(description='csv2sqlite')
    parser.add_argument('-i', '--input_file', required=True,
                        help="Path to CSV records file")
    parser.add_argument('-o', '--output_file', required=True,
                        help="Path to sqlite output file")
    parser.add_argument('--log-level',
                        help="Set logging to specified level")
    return parser.parse_args(args)


def dump2sqlite(records, output_file):
    """Dumps tests results to database."""
    results_keys = records.results[0].keys()
    keys_len = len(results_keys)
    for key in (
            'verdict', 'last_status', 'exported', 'time', 'comment', 'stdout', 'stderr', 'user1'):
        if key not in results_keys:
            results_keys.append(key)

    conn = sqlite3.connect(os.path.expanduser(output_file), detect_types=sqlite3.PARSE_DECLTYPES)

    # in each row there needs to be data for every column
    pad_data = ['' for _ in range(len(results_keys) - keys_len)]
    # last column is current time
    now = datetime.datetime.utcnow()

    def _extend_row(row):
        if pad_data:
            row.extend(pad_data)
        row.append(now)
        return row

    to_db = [_extend_row(row.values()) for row in records.results]

    cur = conn.cursor()
    cur.execute(
        "CREATE TABLE testcases ({},sqltime TIMESTAMP)".format(
            ','.join(['{} TEXT'.format(key) for key in results_keys])))
    cur.executemany("INSERT INTO testcases VALUES ({},?)".format(
        ','.join(['?' for _ in results_keys])), to_db)

    if records.testrun:
        cur.execute("CREATE TABLE testrun (testrun TEXT)")
        cur.execute("INSERT INTO testrun VALUES (?)", (records.testrun, ))

    conn.commit()
    conn.close()

    logger.info("Data written to '{}'".format(output_file))


def main(args=None):
    """Main function for cli."""
    args = get_args(args)

    init_log(args.log_level)

    try:
        records = csvtools.get_imported_data(args.input_file)
    except (EnvironmentError, Dump2PolarionException) as err:
        logger.fatal(err)
        return 1

    # check if all columns required by `pytest_polarion_cfme` are there
    required_columns = {'id': 'ID', 'title': 'Title'}
    missing_columns = [required_columns[k] for k in required_columns if k not in records.results[0]]
    if missing_columns:
        logger.fatal(
            "The input file '{}' is missing following columns: {}".format(
                args.input_file, ', '.join(missing_columns)))
        return 1

    try:
        dump2sqlite(records, args.output_file)
    # pylint: disable=broad-except
    except Exception as err:
        logger.fatal(err)
        return 1

    return 0



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Submit results to the Polarion XUnit/TestCase Importer.
"""

from __future__ import unicode_literals, absolute_import

import os
import io
import logging

import requests

from dump2polarion import msgbus
from dump2polarion.configuration import get_config
from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.utils import xunit_fill_testrun_id

# requests package backwards compatibility mess
# pylint: disable=import-error,ungrouped-imports
from requests.packages.urllib3.exceptions import InsecureRequestWarning as IRWrequests
# pylint: disable=no-member
requests.packages.urllib3.disable_warnings(IRWrequests)
try:
    import urllib3
    from urllib3.exceptions import InsecureRequestWarning as IRWurllib3
    urllib3.disable_warnings(IRWurllib3)
except ImportError:
    pass


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def _get_xml_input(xml_str, xml_file):
    if xml_str:
        xml_input = xml_str
    elif xml_file and os.path.exists(xml_file):
        with io.open(xml_file, encoding='utf-8') as input_file:
            xml_input = input_file.read()
    else:
        raise Dump2PolarionException("Failed to submit results to Polarion - no data supplied")
    return xml_input


def _get_submit_target(xml_input, config):
    if '<testcases' in xml_input:
        target = config.get('testcase_taget')
    elif '<testsuites' in xml_input:
        target = config.get('xunit_target')
    else:
        raise Dump2PolarionException(
            "Failed to submit results to Polarion - submit target not found")
    return target


def _get_credentials(config, **kwargs):
    login = kwargs.get('user') or config.get('username') or os.environ.get("POLARION_USERNAME")
    pwd = kwargs.get('password') or config.get('password') or os.environ.get("POLARION_PASSWORD")

    if not all([login, pwd]):
        raise Dump2PolarionException("Failed to submit results to Polarion - missing credentials")

    return (login, pwd)


def _fill_testrun_id(xml_input, testrun_id):
    if '<testsuites' in xml_input and 'polarion-testrun-id' not in xml_input:
        if not testrun_id:
            raise Dump2PolarionException(
                "Failed to submit results to Polarion - missing testrun id")
        return xunit_fill_testrun_id(xml_input, testrun_id)
    return xml_input


def submit(xml_str=None, xml_file=None, config=None, **kwargs):
    """Submits results to the Polarion Importer."""
    try:
        # get default configuration when missing
        config = config or get_config()
        xml_input = _get_xml_input(xml_str, xml_file)
        credentials = _get_credentials(config, **kwargs)
        submit_target = _get_submit_target(xml_input, config)
        xml_input = _fill_testrun_id(xml_input, kwargs.get('testrun_id'))
    except Dump2PolarionException as err:
        logger.error(err)
        return

    logger.info("Submitting results to {}".format(submit_target))
    files = {'file': ('results.xml', xml_input)}
    try:
        response = requests.post(submit_target, files=files, auth=credentials, verify=False)
    # pylint: disable=broad-except
    except Exception as err:
        logger.error(err)
        response = None

    if response is None:
        logger.error("Failed to submit results to {}".format(submit_target))
    elif response:
        logger.info("Results received by the Importer (HTTP status {})".format(
            response.status_code))
    else:
        logger.error("HTTP status {}: failed to submit results to {}".format(
            response.status_code, submit_target))

    return response


def submit_and_verify(xml_str=None, xml_file=None, config=None, **kwargs):
    """Submits results to the Polarion Importer and checks that it was imported."""
    try:
        xml_input = _get_xml_input(xml_str, xml_file)
    except Dump2PolarionException as err:
        logger.error(err)
        return

    # get default configuration when missing
    config = config or get_config()

    if kwargs.get('no_verify'):
        verification_func = None
    else:
        msgbus_login = kwargs.get('msgbus_user') or config.get('msgbus_username') or kwargs.get(
            'user') or config.get('username') or os.environ.get("POLARION_USERNAME")
        msgbus_pwd = kwargs.get('msgbus_password') or config.get('msgbus_password') or kwargs.get(
            'password') or config.get('password') or os.environ.get("POLARION_PASSWORD")

        verification_func = msgbus.get_verification_func(
            config.get('message_bus'),
            xml_input,
            user=msgbus_login,
            password=msgbus_pwd,
            log_file=kwargs.get('msgbus_log'))

    response = submit(xml_input, config=config, **kwargs)

    if verification_func:
        response = verification_func(skip=not response, timeout=kwargs.get('verify_timeout'))

    return bool(response)



# -*- coding: utf-8 -*-

import csv
import codecs


class UTF8Recoder(object):
    """Iterator that reads an encoded stream and reencodes the input to UTF-8."""
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf-8')


class UnicodeReader(object):
    """ A CSV reader which will iterate over lines in the CSV file that
    is encoded in the given encoding."""
    def __init__(self, f, dialect=csv.excel, encoding='utf-8', **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, 'utf-8') for s in row]

    def __iter__(self):
        return self



# -*- coding: utf-8 -*-
"""
Allowed values for verdicts.
"""

from __future__ import unicode_literals


class Verdicts(object):
    PASS = ('passed', 'pass')
    FAIL = ('failed', 'fail')
    SKIP = ('skipped', 'skip', 'blocked')
    WAIT = ('null', 'wait', 'waiting')



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Dump testcases results to xunit file for submitting to the Polarion XUnit Importer.
"""

from __future__ import unicode_literals, absolute_import

import datetime
import string
import random

from collections import namedtuple

from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement, Comment

from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.verdicts import Verdicts
from dump2polarion.transform import get_results_transform
from dump2polarion.utils import write_xml, get_unicode_str


ImportedData = namedtuple('ImportedData', 'results testrun')


class XunitExport(object):
    """Exports testcases results into Polarion xunit."""
    def __init__(self, testrun_id, tests_records, config, transform_func=None):
        self.testrun_id = testrun_id
        self.tests_records = tests_records
        self.config = config
        self._lookup_prop = ''
        self._transform_func = transform_func or get_results_transform(config)

    def _top_element(self):
        """Returns top XML element."""
        top = Element('testsuites')
        comment = Comment("Generated for testrun {}".format(self.testrun_id))
        top.append(comment)
        return top

    def _properties_element(self, parent_element):
        """Returns properties XML element."""
        testsuites_properties = SubElement(parent_element, 'properties')

        SubElement(testsuites_properties, 'property',
                   {'name': 'polarion-testrun-id', 'value': str(self.testrun_id)})

        response_prop_set = False
        for name, value in self.config['xunit_import_properties'].iteritems():
            SubElement(testsuites_properties, 'property',
                       {'name': name, 'value': str(value)})
            if 'polarion-response-' in name:
                response_prop_set = True
            elif name == 'polarion-lookup-method':
                self._lookup_prop = str(value)

        if not response_prop_set:
            name = 'polarion-response-dump2polarion'
            value = ''.join(random.sample(string.lowercase, 10)) + '8'
            SubElement(testsuites_properties, 'property', {'name': name, 'value': value})

        if not self._lookup_prop:
            if 'id' in self.tests_records.results[0]:
                self._lookup_prop = 'ID'
            elif 'title' in self.tests_records.results[0]:
                self._lookup_prop = 'Name'
            else:
                raise Dump2PolarionException(
                    "Failed to set the 'polarion-lookup-method' property")
            SubElement(testsuites_properties, 'property',
                       {'name': 'polarion-lookup-method', 'value': self._lookup_prop})
        elif self._lookup_prop.lower() not in ('id', 'name', 'custom'):
            raise Dump2PolarionException(
                "Invalid value '{}' for the 'polarion-lookup-method' property".format(
                    self._lookup_prop))

        return testsuites_properties

    def _testsuite_element(self, parent_element):
        """Returns testsuite XML element."""
        testsuite = SubElement(
            parent_element,
            'testsuite',
            {'name': 'Import for {} - {} testrun'.format(
                self.config['xunit_import_properties']['polarion-project-id'], self.testrun_id)})
        return testsuite

    @staticmethod
    def _fill_verdict(verdict, result, testcase, records):
        # xunit Pass maps to Passed in Polarion
        if verdict in Verdicts.PASS:
            records['passed'] += 1
        # xunit Failure maps to Failed in Polarion
        elif verdict in Verdicts.FAIL:
            records['failures'] += 1
            verdict_data = {'type': 'failure'}
            if result.get('comment'):
                verdict_data['message'] = get_unicode_str(result['comment'])
            SubElement(testcase, 'failure', verdict_data)
        # xunit Error maps to Blocked in Polarion
        elif verdict in Verdicts.SKIP:
            records['skipped'] += 1
            verdict_data = {'type': 'error'}
            if result.get('comment'):
                verdict_data['message'] = get_unicode_str(result['comment'])
            SubElement(testcase, 'error', verdict_data)
        # xunit Skipped maps to Waiting in Polarion
        elif verdict in Verdicts.WAIT:
            records['waiting'] += 1
            verdict_data = {'type': 'skipped'}
            if result.get('comment'):
                verdict_data['message'] = get_unicode_str(result['comment'])
            SubElement(testcase, 'skipped', verdict_data)

    def _gen_testcase(self, parent_element, result, records):
        """Creates XML element for given testcase result and update testcases records."""
        if self._transform_func:
            result = self._transform_func(result)
            if not result:
                return
        verdict = result.get('verdict', '').strip().lower()
        if verdict not in Verdicts.PASS + Verdicts.FAIL + Verdicts.SKIP + Verdicts.WAIT:
            return
        testcase_id = result.get('id')
        testcase_title = result.get('title')
        if not testcase_id and self._lookup_prop.lower() == 'id':
            return
        if not testcase_title and self._lookup_prop.lower() == 'name':
            return

        testcase_time = float(result.get('time') or result.get('duration') or 0)
        records['time'] += testcase_time

        testcase_data = {
            'name': testcase_title or testcase_id,
            'time': str(testcase_time)}
        if result.get('classname'):
            testcase_data['classname'] = result['classname']
        testcase = SubElement(parent_element, 'testcase', testcase_data)

        self._fill_verdict(verdict, result, testcase, records)

        if result.get('stdout'):
            system_out = SubElement(testcase, 'system-out')
            system_out.text = get_unicode_str(result['stdout'])

        if result.get('stderr'):
            system_err = SubElement(testcase, 'system-err')
            system_err.text = get_unicode_str(result['stderr'])

        properties = SubElement(testcase, 'properties')
        SubElement(properties, 'property',
                   {'name': 'polarion-testcase-id', 'value': testcase_id or testcase_title})
        if verdict in Verdicts.PASS and result.get('comment'):
            SubElement(properties, 'property',
                       {'name': 'polarion-testcase-comment',
                        'value': get_unicode_str(result['comment'])})

    def _fill_tests_results(self, testsuite_element):
        """Creates records for all testcases results."""
        if not self.tests_records.results:
            raise Dump2PolarionException("Nothing to export")

        records = dict(passed=0, skipped=0, failures=0, waiting=0, time=0.0)
        for testcase_result in self.tests_records.results:
            self._gen_testcase(testsuite_element, testcase_result, records)

        tests_num = (
            records['passed'] +
            records['skipped'] +
            records['failures'] +
            records['waiting'])

        if tests_num == 0:
            raise Dump2PolarionException("Nothing to export")

        testsuite_element.set('errors', str(records['skipped']))
        testsuite_element.set('failures', str(records['failures']))
        testsuite_element.set('skipped', str(records['waiting']))
        testsuite_element.set('time', str(records['time']))
        testsuite_element.set('tests', str(tests_num))

    @staticmethod
    def _prettify(top_element):
        """Returns a pretty-printed XML."""
        rough_string = ElementTree.tostring(top_element, 'utf-8')
        reparsed = minidom.parseString(rough_string)
        return get_unicode_str(reparsed.toprettyxml(indent='  ', encoding='utf-8'))

    def export(self):
        """Returns xunit XML."""
        top = self._top_element()
        self._properties_element(top)
        testsuite = self._testsuite_element(top)
        self._fill_tests_results(testsuite)
        return self._prettify(top)

    def write_xml(self, xml, output_file=None):
        """Outputs the XML content into a file."""
        gen_filename = 'testrun_{}-{:%Y%m%d%H%M%S}.xml'.format(
            self.testrun_id, datetime.datetime.now())
        write_xml(xml, output_loc=output_file, filename=gen_filename)



# -*- coding: utf-8 -*-
"""
Helper functions for handling data in sqlite3.
"""

from __future__ import unicode_literals, absolute_import

import os

import logging
import sqlite3
from sqlite3 import Error as SQLiteError

from collections import OrderedDict

from dump2polarion.exporter import ImportedData
from dump2polarion.exceptions import Dump2PolarionException


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


SQLITE_EXT = ('.sqlite', '.sqlite3', '.db', '.db3')


def _get_testrun_from_sqlite(conn):
    """Returns testrun id saved from original csv file."""
    cur = conn.cursor()
    try:
        cur.execute('SELECT testrun FROM testrun')
        return cur.fetchone()[0]
    except (IndexError, SQLiteError):
        return


def _open_sqlite(db_file):
    """Opens database connection."""
    db_file = os.path.expanduser(db_file)
    try:
        with open(db_file):
            # test that the file can be accessed
            pass
        return sqlite3.connect(db_file, detect_types=sqlite3.PARSE_DECLTYPES)
    except (IOError, SQLiteError) as err:
        raise Dump2PolarionException('{}'.format(err))


# pylint: disable=unused-argument
def import_sqlite(db_file, older_than=None, **kwargs):
    """Reads the content of the database file and returns imported data."""
    conn = _open_sqlite(db_file)
    cur = conn.cursor()
    # get rows that were not exported yet
    select = "SELECT * FROM testcases WHERE exported != 'yes'"
    if older_than:
        cur.execute(' '.join((select, "AND sqltime < ?")), (older_than, ))
    else:
        cur.execute(select)
    columns = [description[0] for description in cur.description]
    rows = cur.fetchall()

    # map data to columns
    results = []
    for row in rows:
        record = OrderedDict(zip(columns, row))
        results.append(record)

    testrun = _get_testrun_from_sqlite(conn)

    conn.close()

    return ImportedData(results=results, testrun=testrun)


def mark_exported_sqlite(db_file, older_than=None):
    """Marks rows with verdict as exported."""
    logger.debug("Marking rows in database as exported")
    conn = _open_sqlite(db_file)
    cur = conn.cursor()
    update = "UPDATE testcases SET exported = 'yes' WHERE verdict IS NOT null AND verdict != ''"
    if older_than:
        cur.execute(' '.join((update, "AND sqltime < ?")), (older_than, ))
    else:
        cur.execute(update)
    conn.commit()
    conn.close()



from setuptools import setup, find_packages

setup(
    name='dump2polarion',
    version='0.8',
    url='https://github.com/mkoura/dump2polarion',
    description='Dump testcases results to xunit file and submit it to Polarion xunit importer',
    long_description=open('README.rst').read().strip(),
    author='Martin Kourim',
    author_email='mkourim@redhat.com',
    license='GPL',
    packages=find_packages(exclude=('tests',)),
    scripts=['csv2sqlite.py', 'polarion_dumper.py'],
    install_requires=['requests', 'pyyaml', 'stomp.py'],
    keywords=['polarion', 'testing'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 2.7'],
    include_package_data=True
)



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Dump testcases results from a CSV, SQLite, junit or Ostriz input file to xunit file and submit it
to the Polarion XUnit Importer.
"""

from __future__ import unicode_literals, absolute_import

import os
import io
import argparse
import logging
import datetime

import dump2polarion
from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.utils import init_log
from dump2polarion import dbtools


# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def get_args(args=None):
    """Get command line arguments."""
    parser = argparse.ArgumentParser(description='dump2polarion')
    parser.add_argument('-i', '--input_file', required=True,
                        help="Path to CSV or SQLite reports file or xunit XML file")
    parser.add_argument('-o', '--output_file',
                        help="Where to save the XML output file (default: not saved)")
    parser.add_argument('-t', '--testrun-id',
                        help="Polarion test run id")
    parser.add_argument('-c', '--config-file',
                        help="Path to config YAML")
    parser.add_argument('-n', '--no-submit', action='store_true',
                        help="Don't submit results to Polarion")
    parser.add_argument('--user',
                        help="Username to use to submit results to Polarion")
    parser.add_argument('--password',
                        help="Password to use to submit results to Polarion")
    parser.add_argument('--msgbus-user',
                        help="Username to use to connect to the message bus")
    parser.add_argument('--msgbus-password',
                        help="Password to use to connect to the message bus")
    parser.add_argument('-f', '--force', action='store_true',
                        help="Don't validate test run id")
    parser.add_argument('--no-verify', action='store_true',
                        help="Don't verify results submission")
    parser.add_argument('--verify-timeout', type=int, default=300, metavar='SEC',
                        help="How long to wait (in seconds) for verification of results submission"
                             " (default: %(default)s)")
    parser.add_argument('--msgbus-log',
                        help="Where to save the log file returned by msgbus (default: not saved)")
    parser.add_argument('--log-level',
                        help="Set logging to specified level")
    return parser.parse_args(args)


def get_testrun_id(args, testrun_id):
    """Returns testrun id."""
    if (args.testrun_id and testrun_id and not args.force and
            testrun_id != args.testrun_id):
        raise Dump2PolarionException(
            "The test run id '{}' found in exported data doesn't match '{}'. "
            "If you really want to proceed, add '-f'.".format(testrun_id, args.testrun_id))

    found_testrun_id = args.testrun_id or testrun_id
    if not found_testrun_id:
        raise Dump2PolarionException(
            "The testrun id was not specified on command line and not found in the input data.")

    return found_testrun_id


def submit_if_ready(args, config):
    """Submits the input XML file if it's already in the expected format."""
    _, ext = os.path.splitext(args.input_file)
    if ext.lower() != '.xml':
        return

    with io.open(args.input_file, encoding='utf-8') as input_file:
        xml = input_file.read()

    if '<testsuites' in xml or '<testcases' in xml:
        if args.no_submit:
            logger.info("Nothing to do")
            return 0
        # expect importer xml and just submit it
        response = dump2polarion.submit_and_verify(xml, config=config, **vars(args))
        return 0 if response else 2


def main(args=None):
    """Main function for cli."""
    args = get_args(args)

    init_log(args.log_level)

    try:
        config = dump2polarion.get_config(args.config_file)
    except Dump2PolarionException as err:
        logger.fatal(err)
        return 1

    submit_outcome = submit_if_ready(args, config)
    if submit_outcome is not None:
        # submitted, nothing more to do
        return submit_outcome

    import_time = datetime.datetime.utcnow()

    try:
        records = dump2polarion.do_import(args.input_file, older_than=import_time)
        testrun_id = get_testrun_id(args, records.testrun)
        exporter = dump2polarion.XunitExport(testrun_id, records, config)
        output = exporter.export()
    except (EnvironmentError, Dump2PolarionException) as err:
        logger.fatal(err)
        return 1

    if args.output_file or args.no_submit:
        # when no output file is specified, the 'testrun_TESTRUN_ID-TIMESTAMP'
        # file will be created in current directory
        exporter.write_xml(output, args.output_file)

    if not args.no_submit:
        response = dump2polarion.submit_and_verify(output, config=config, **vars(args))

        _, ext = os.path.splitext(args.input_file)
        if ext.lower() in dbtools.SQLITE_EXT and response:
            dbtools.mark_exported_sqlite(args.input_file, import_time)

        return 0 if response else 2

    return 0



# -*- coding: utf-8 -*-
"""
Import data using correct tools.
"""

from __future__ import unicode_literals, absolute_import

import os

from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.dbtools import SQLITE_EXT


def _get_importer(input_file):
    """Selects importer based on input file type."""
    _, ext = os.path.splitext(input_file)
    ext = ext.lower()

    if 'ostriz' in input_file:
        from dump2polarion import ostriztools
        importer = ostriztools.import_ostriz
    elif ext == '.xml':
        # expect junit-report from pytest
        from dump2polarion import junittools
        importer = junittools.import_junit
    elif ext == '.csv':
        from dump2polarion import csvtools
        importer = csvtools.import_csv
    elif ext in SQLITE_EXT:
        from dump2polarion import dbtools
        importer = dbtools.import_sqlite
    else:
        raise Dump2PolarionException(
            "Cannot recognize type of input data, add file extension.")

    return importer


def do_import(input_file, **kwargs):
    """Imports the input file."""

    return _get_importer(input_file)(input_file, **kwargs)



# -*- coding: utf-8 -*-
"""
Imports.
"""

from dump2polarion.exporter import XunitExport
from dump2polarion.importer import do_import
from dump2polarion.configuration import get_config
from dump2polarion.submit import submit_and_verify


__all__ = ['XunitExport', 'do_import', 'get_config', 'submit_and_verify']



# -*- coding: utf-8 -*-
# pylint: disable=logging-format-interpolation
"""
Helper functions for handling data in pytest junit format.
"""

from __future__ import unicode_literals, absolute_import

import os

from collections import OrderedDict

from xml.etree import ElementTree

from dump2polarion.exporter import ImportedData
from dump2polarion.exceptions import Dump2PolarionException


# pylint: disable=unused-argument
def import_junit(junit_file, **kwargs):
    """Reads the content of the junit-results file produced by pytest and returns imported data."""
    try:
        tree = ElementTree.parse(os.path.expanduser(junit_file))
    except Exception as err:
        raise Dump2PolarionException("Failed to parse XML file '{}': {}".format(junit_file, err))
    xml_root = tree.getroot()

    results = []
    for test_data in xml_root:
        if test_data.tag != 'testcase':
            continue

        verdict = 'failed'
        comment = ''
        for element in test_data:
            # we don't want to submit errors and failures anyway
            if element.tag in ('error', 'failure'):
                break
            elif element.tag == 'skipped':
                comment = element.get('message')
                if comment:
                    verdict = 'skipped'
                break
        else:
            verdict = 'passed'

        if verdict == 'failed':
            continue

        title = test_data.get('name')
        classname = test_data.get('classname')
        time = test_data.get('time', 0)

        record = OrderedDict([
            ('title', title),
            ('classname', classname),
            ('verdict', verdict),
            ('comment', comment),
            ('time', time)])
        results.append(record)

    return ImportedData(results=results, testrun=None)



# -*- coding: utf-8 -*-
"""
Functions for validating and transforming results. These are specific for given Polarion project.

If the 'polarion-lookup-method' is set to 'custom', this is the place where you can
set the 'id' of the testcase to desired value.
"""

from __future__ import unicode_literals, absolute_import

import re

from dump2polarion.verdicts import Verdicts


# pylint: disable=unused-argument
def get_results_transform_cfme(config):
    """Return result transformation function for CFME."""
    cfme_searches = [
        'Skipping due to these blockers',
        'BZ ?[0-9]+',
        'GH ?#?[0-9]+',
        'GH#ManageIQ',
    ]
    cfme_skips = re.compile('(' + ')|('.join(cfme_searches) + ')')

    def results_transform_cfme(result):
        """Results transform for CFME."""
        if result.get('classname'):
            # we don't need classnames?
            del result['classname']
        verdict = result.get('verdict', '').strip().lower()
        # we want to submit PASS and WAIT results
        if verdict in Verdicts.PASS + Verdicts.WAIT:
            return result
        # ... and SKIP results where there is a good reason (blocker etc.)
        if verdict in Verdicts.SKIP:
            comment = result.get('comment')
            if comment and cfme_skips.search(comment):
                # found reason for skip
                return result

    return results_transform_cfme


PROJECT_MAPPING = {
    'RHCF3': get_results_transform_cfme,
}


def get_results_transform(config):
    """Returns results transformation function.

    The transformation function is returned by calling corresponding "getter" function.

    This allows customizations of results data according to requirements
    of the specific project.

    When no results data are returned, this result will be ignored
    and will not be written to the resulting XML.
    """

    project = config['xunit_import_properties']['polarion-project-id']
    if project in PROJECT_MAPPING:
        return PROJECT_MAPPING[project](config)
